@@ -1,7 +1,7 @@
 ;;; iiimp.el --- IIIM Protocol Library
 ;;;                   
 
-;; Copyright (C) 2000 MIYASHITA Hisashi
+;; Copyright (C) 2000-2001 MIYASHITA Hisashi
 
 ;; Author: MIYASHITA Hisashi <himi@m17n.org>
 ;;
@@ -34,7 +34,7 @@
   (defvar iiimp-debug-flag nil
     "IIIMP debug mode flag."))
 
-(defvar iiimp-protocol-version 2
+(defconst iiimp-protocol-version 2
   "IIIMP version number.")
 
 (defvar iiimp-debug-buffer nil)
@@ -320,6 +320,14 @@
   (aset com-id 4 pt)
   com-id)
 
+(defun iiimp-proc-to-com-id (proc)
+  (let ((ll iiimp-communication-list))
+    (while
+	(and ll
+	     (not (iiimp-com-id-process (car ll))))
+      (setq ll (cdr ll)))
+    (car ll)))
+
 ;; IM-ID := (COM-ID id._NUMBER_of_IM)
 
 (defsubst iiimp-im-id-to-com-id (im-id)
@@ -344,6 +352,17 @@
 
 ;;; network I/O
 
+(defsubst iiimp-check-channel-connection (com-id)
+  (eq (process-status (iiimp-com-id-process com-id)) 'open))
+
+(defun iiimp-network-sentinel (proc stat)
+  (let ((com-id (iiimp-proc-to-com-id proc)))
+    (if (and com-id
+	     (not (iiimp-check-channel-connection com-id)))
+	(progn
+	  (iiimp-disable-async-invocation com-id)
+	  (iiimp-connection-error "sentinel" com-id)))))
+
 (defun iiimp-open-network-channel (host port buf)
   (let ((coding-system-for-read nil)
 	(coding-system-for-write nil)
@@ -357,6 +376,8 @@
 	    (open-network-stream iiimp-process-name buf host port))
       (process-kill-without-query proc)
       (set-process-coding-system proc 'binary 'binary)
+      (set-process-sentinel
+       proc (function iiimp-network-sentinel))
       (set-marker-insertion-type (process-mark proc) t))
     proc))
 
@@ -378,9 +399,10 @@
 	 (stat (process-status proc)))
     (while (not (or (eq stat 'closed)
 		    (eq stat nil)))
-	(delete-process proc)
-	(setq stat (process-status proc))
-	(sleep-for 1))
+      (set-process-sentinel proc nil)
+      (delete-process proc)
+      (setq stat (process-status proc))
+      (sleep-for 1))
     (kill-buffer (iiimp-com-id-buffer com-id))
     (setq iiimp-communication-list
 	  (delq com-id iiimp-communication-list))
@@ -404,6 +426,16 @@
 		 iiimp-async-invocation-handler-1)
 		nil t))))
 
+(defun iiimp-disable-async-invocation (com-id)
+  (let ((buf (iiimp-com-id-buffer com-id)))
+    (if (buffer-live-p buf)
+	(save-excursion
+	  (set-buffer buf)
+	  (remove-hook 'after-change-functions
+		       (function
+			iiimp-async-invocation-handler-1)
+		       t)))))
+
 ;;
 ;;
 ;;
@@ -414,15 +446,26 @@
      (let ((after-change-functions nil))
        ,@form)))
 
-(defsubst iiimp-check-channel-connection (com-id)
-  (eq (process-status (iiimp-com-id-process com-id)) 'open))
-
 (defun iiimp-check-channel-bytes (com-id start bytes wait)
-  (if (null wait) (>= (point-max) (+ start bytes))
-    (if (not (numberp wait)) (setq wait nil))
-    (while (< (point-max) (+ start bytes))
-      (accept-process-output (iiimp-com-id-process com-id) wait))
-    t))
+  (cond ((eq wait t)
+	 (catch 'tag
+	   (while (< (point-max) (+ start bytes))
+	     (if (iiimp-check-channel-connection com-id)
+		 (accept-process-output
+		  (iiimp-com-id-process com-id)
+		  1)
+	       (throw 'tag nil)))
+	   t))
+	((numberp wait)
+	 (if (< (point-max) (+ start bytes))
+	     (accept-process-output
+	      (iiimp-com-id-process com-id)
+	      wait))
+	 (>= (point-max) (+ start bytes)))
+	((null wait)
+	 (>= (point-max) (+ start bytes)))
+	(t
+	 (error "Invalid argument(wait):%S" wait))))
 
 (defun iiimp-wait-channel (com-id until)
   (while (< (point-max) until)
